#include <string.h>
#include <ctype.h>

#include <iostream.h>

#include "hist.h"
#include "util.h"

FitsHist::FitsHist(FitsFile* fits, int width, int height,
		   Matrix& n, Function func, float block)
{
  // we need to translate by .5 to center the pixel
  // needed for binning

  Matrix m = n * Translate(.5,.5);
  //  Matrix m = n;

  // we need to translate by another .5 for the offset from Image 
  // to Data coordinates need for LTM/V and WCS

  Matrix mm = n * Translate(1,1);

  FitsHead* srcHead = fits->head();
  FitsTableHDU* srcHDU = (FitsTableHDU*)(srcHead->hdu());

  int size = width*height;

  // make sure we have a table with columns, X, Y

  if (!fits->isTable())
    return;

  // get X column

  FitsColumn* xcol = NULL;
  if (fits->pBinX())
    xcol = srcHDU->find(fits->pBinX());

  if (!xcol)
    return;

  // get Y column

  FitsColumn* ycol = NULL;
  if (fits->pBinY())
    ycol = srcHDU->find(fits->pBinY());

  if (!ycol)
    return;

  // get Z column (if specified)

  FitsColumn* zcol = NULL;
  if (fits->pBinZ())
    zcol = srcHDU->find(fits->pBinZ());

  // create header

  head_ = new FitsHead(width, height, -32);
  if (!head_->isValid())
    return;

  // append other keywords


  // LTM/LTV

  head_->appendReal("LTM1_1", mm[0][0], 10, NULL);
  head_->appendReal("LTM1_2", mm[0][1], 10, NULL);
  head_->appendReal("LTM2_1", mm[1][0], 10, NULL);
  head_->appendReal("LTM2_2", mm[1][1], 10, NULL);
  head_->appendReal("LTV1"  , mm[2][0], 10, NULL);
  head_->appendReal("LTV2"  , mm[2][1], 10, NULL);

  // OBJECT

  char* object = srcHead->getString("OBJECT");
  if (object) {
    head_->appendString("OBJECT", object, NULL);
    delete object;
  }

  // WCS

  char* ctype1 = srcHead->getString(ftcat("TCTYP", xcol->index()));
  char* ctype2 = srcHead->getString(ftcat("TCTYP", ycol->index()));
  float crval1 = srcHead->getReal(ftcat("TCRVL", xcol->index()), 0);
  float crval2 = srcHead->getReal(ftcat("TCRVL", ycol->index()), 0);
  float crpix1 = srcHead->getReal(ftcat("TCRPX", xcol->index()), 0);
  float crpix2 = srcHead->getReal(ftcat("TCRPX", ycol->index()), 0);
  float cdelt1 = srcHead->getReal(ftcat("TCDLT", xcol->index()), 0);
  float cdelt2 = srcHead->getReal(ftcat("TCDLT", ycol->index()), 0);

  char* radecsys = srcHead->getString("RADECSYS");
  float equinox = srcHead->getReal("EQUINOX", 0);

  // this is to provide backward compatibility

  if (!(ctype1 && ctype2)) {
    ctype1 = srcHead->getString(ftcat("CTYPE", 1));
    ctype2 = srcHead->getString(ftcat("CTYPE", 2));
    crval1 = srcHead->getReal(ftcat("CRVAL", 1), 0);
    crval2 = srcHead->getReal(ftcat("CRVAL", 2), 0);
    crpix1 = srcHead->getReal(ftcat("CRPIX", 1), 0);
    crpix2 = srcHead->getReal(ftcat("CRPIX", 2), 0);
    cdelt1 = srcHead->getReal(ftcat("CDELT", 1), 0);
    cdelt2 = srcHead->getReal(ftcat("CDELT", 2), 0);
  }

  // xlate crpix * binToData * dataToImage

  Vector pix = Vector(crpix1,crpix2) * mm;
  Vector delt = Vector(cdelt1,cdelt2) * block;

  if (ctype1 && ctype2) {
    head_->appendString("CTYPE1", ctype1, NULL);
    head_->appendString("CTYPE2", ctype2, NULL);
    head_->appendReal("CRVAL1", crval1, 10, NULL);
    head_->appendReal("CRVAL2", crval2, 10, NULL);
    head_->appendReal("CRPIX1", pix[0], 10, NULL);
    head_->appendReal("CRPIX2", pix[1], 10, NULL);
    head_->appendReal("CDELT1", delt[0], 10, NULL);
    head_->appendReal("CDELT2", delt[1], 10, NULL);
    if (radecsys)
      head_->appendString("RADECSYS", radecsys, NULL);
    if (!equinox)
      head_->appendReal("EQUINOX", equinox, 10, NULL);
  }

  if (ctype1)
    delete ctype1;
  if (ctype2)
    delete ctype2;
  if (radecsys)
    delete radecsys;

  // create image space

  float* dest = new float[size];
  memset(dest, 0, size*sizeof(float));

  // bin it up

  char* ptr = (char*)fits->data();
  int rowlen = srcHDU->width();
  int rows = srcHDU->rows();

  register double m00 = m.matrix(0,0);
  register double m10 = m.matrix(1,0);
  register double m20 = m.matrix(2,0);
  register double m01 = m.matrix(0,1);
  register double m11 = m.matrix(1,1);
  register double m21 = m.matrix(2,1);

  // unroll the loops to speed things up

  if (func==SUM) { // SUM Function
    if (zcol) { // bin on third column
      for (int i=0; i<rows; i++, ptr+=rowlen) {
	register double x = xcol->value(ptr);
	register double y = ycol->value(ptr);

	register double X = x*m00 + y*m10 + m20;
	register double Y = x*m01 + y*m11 + m21;

	if (X >= 0 && X < width && Y >= 0 && Y < height)
	  dest[((int)Y)*width + (int)X]+=zcol->value(ptr);
      }
    }
    else { // bin on counts
      for (int i=0; i<rows; i++, ptr+=rowlen) {
	register double x = xcol->value(ptr);
	register double y = ycol->value(ptr);

	register double X = x*m00 + y*m10 + m20;
	register double Y = x*m01 + y*m11 + m21;

	if (X >= 0 && X < width && Y >= 0 && Y < height)
	  dest[((int)Y)*width + (int)X]++;
      }
    }
  }
  else {  // AVERAGE function
    if (zcol) { // bin on third column
      int* counts = new int[size];
      memset(counts, 0, size*sizeof(int));

      for (int i=0; i<rows; i++, ptr+=rowlen) {
	register double x = xcol->value(ptr);
	register double y = ycol->value(ptr);

	register double X = x*m00 + y*m10 + m20;
	register double Y = x*m01 + y*m11 + m21;

	if (X >= 0 && X < width && Y >= 0 && Y < height) {
	  int index = ((int)Y)*width + (int)X;
	  dest[index]+=zcol->value(ptr);
	  counts[index]++;
	}
      }

      for (int k=0; k<size; k++)
	if (counts[k]>0)
	  dest[k] /= counts[k];
	else
	  dest[k] = 0;
    }
    else { // bin on counts
      for (int i=0; i<rows; i++, ptr+=rowlen) {
	register double x = xcol->value(ptr);
	register double y = ycol->value(ptr);

	register double X = x*m00 + y*m10 + m20;
	register double Y = x*m01 + y*m11 + m21;

	if (X >= 0 && X < width && Y >= 0 && Y < height)
	  dest[((int)Y)*width + (int)X]++;
      }

      for (int k=0; k<size; k++)
	dest[k] /= block;
    }
  }
    
  data_ = dest;
  byteswap_ = 0; // we've already swapped the data during binning
  valid_ = 1;
}

FitsHist::~FitsHist()
{
  if (data_)
    delete (float*)data_;
}
