#include "rectangle.h"
#include "frame.h"
#include "util.h"

Rectangle::Rectangle(Frame* p, const Vector& ctr, const Vector& s, double a,
		     const char* clr, const char* f, const char* t, 
		     unsigned short prop)
  : AnnuliMarker(p, ctr, 4, clr, f, t, prop)
{
  size = s;
  angle = a;
  strcpy(type,"rectangle");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Rectangle::updateCoords(Matrix& m)
{
  size *= Scale(m);
  Marker::updateCoords(m);
}

void Rectangle::editBegin(int h)
{
  switch (h) {
  case 1:
    return;
  case 2:
    size = Vector(-size[0],size[1]);
    return;
  case 3:
    size = -size;
    return;
  case 4:
    size = Vector(size[0],-size[1]);
    return;
  }
}

void Rectangle::edit(const Vector& v, int h)
{
  // This sizes about the center
  //  size = ((v-center)*Rotate(angle)*2).abs();

  // This sizes about the opposite node
  Vector op = center+size/2;
  size = op-(v*Translate(-center)*Rotate(angle)*Translate(center));
  center = v+(size/2)*Rotate(-angle);

  updateBBox();
  doCallBacks(&editCB);
}

void Rectangle::editEnd()
{
  size = size.abs();
  updateBBox();
}

void Rectangle::rotate(const Vector& v, int h)
{
  // v is in ref coords
  // handles are in canvas coords

  Vector a = v - center;
  Vector b = (handle[h-1] * parent->canvasToRef) - center;

  angle += a.angle() - b.angle();

  updateBBox();
  doCallBacks(&rotateCB);
}

void Rectangle::setSize(const Vector& s)
{
  size = s;
  updateBBox();

  doCallBacks(&editCB);
}

void Rectangle::setAngle(double a)
{
  angle = a;
  updateBBox();

  doCallBacks(&rotateCB);
}

int Rectangle::isIn(const Vector& v)
{
  // v is in canvas coords

  Vector s = size * maxAnnuli();
  if (properties & FIXED)
    s /= parent->getZoom();

  Vector p = v * parent->canvasToRef * Translate(-center) * Rotate(angle);

  if ( p[0] < -s[0]/2 || p[0] > s[0]/2 || p[1] < -s[1]/2 || p[1] > s[1]/2)
    return 0;
  else
    return 1;
}

// private

void Rectangle::calcPoints(double zoom)
{
  Matrix m;
  if (properties & FIXED)
    m = Scale(1/zoom) * Rotate(-angle) * Translate(center);
  else
    m = Rotate(-angle) * Translate(center);

  for (int i=0; i<annuli; i++) {
    Matrix mm = Scale(ratio[i]) * m;
    vertices[i][0] = Vector(-size[0]/2.,-size[1]/2.) * mm;
    vertices[i][1] = Vector( size[0]/2.,-size[1]/2.) * mm;
    vertices[i][2] = Vector( size[0]/2., size[1]/2.) * mm;
    vertices[i][3] = Vector(-size[0]/2., size[1]/2.) * mm;
  }
}

void Rectangle::updateHandles()
{
  Vector s = size * maxAnnuli();
  if (properties & FIXED)
    s /= parent->getZoom();

  Matrix m = Rotate(-angle) * Translate(center) * parent->refToCanvas;
  handle[0] = Vector(-s[0]/2.,-s[1]/2.) * m;
  handle[1] = Vector( s[0]/2.,-s[1]/2.) * m;
  handle[2] = Vector( s[0]/2., s[1]/2.) * m;
  handle[3] = Vector(-s[0]/2., s[1]/2.) * m;
}

// list

void Rectangle::list(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector s = parent->mapDistFromRef(size, 
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << setprecision(8) << v << s << radToDeg(angle);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector s = parent->mapDistFromRef(size, 
				      (Frame::CoordSystem)sys,Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << setprecision(8) << v << s[0] << "\" " << s[1] << "\" "
	    << radToDeg(angle);
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << ' ' << ra << ' ' << dec << ' ' 
	    << s[0] << "\" " << s[1] << "\" "
	    << radToDeg(angle);
      }
      break;
    }
    break;
  }

  listProperties(str);
}

void Rectangle::listSAOtng(ostream& str, int sys, int format)
{
  if (text[0])
    str << '#' << text << endl;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << "box(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << size[0] << ',' << size[1] << ',' << radToDeg(angle) << ") ";
    }
    break;
  case Frame::PHYSICAL:
    {
      Vector s = parent->mapDistFromRef(size, 
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << "box(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ") ";
    }
    break;
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << "box(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	    << size[0] << ',' << size[1] << ',' << radToDeg(angle) << ") ";
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << "box(" << ra << ',' << dec << ',' 
	    << size[0] << ',' << size[1] << ',' << radToDeg(angle) << ") ";
      }
      break;
    }
    break;
  }

  str << "# " << colorName;
}

void Rectangle::listPros(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector s = parent->mapDistFromRef(size, 
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << "rotbox " << setprecision(8) << v << s << radToDeg(angle);
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector s = parent->mapDistFromRef(size, 
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << "rotbox " << setprecision(8) << v << s << radToDeg(angle);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    parent->listProsCoordSystem(str, (Frame::CoordSystem)sys);

    // display wcs radius in ARCSEC
    Vector s = parent->mapDistFromRef(size, 
				      (Frame::CoordSystem)sys,Frame::ARCSEC);

    str << "; ";
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << "rotbox " << setprecision(8) << v[0] << "d " << v[1] << "d " 
	    << s[0] << "\" " << s[1] << "\" " << radToDeg(angle);
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << "rotbox " << ra << ' ' << dec+1 << ' '
	      << s[0] << "\" " << s[1] << "\" " << radToDeg(angle);
	else
	  str << "rotbox " << ra << ' ' << dec << ' '
	      << s[0] << "\" " << s[1] << "\" " << radToDeg(angle);
      }
      break;
    }
    break;
  }
}

void Rectangle::listSAOimage(ostream& str)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << "box(" << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << size[0] << ',' << size[1] << ") ";
}
